/**
 * Copyright (C) 2010 Cloudfarming <info@cloudfarming.nl>
 *
 * Licensed under the Eclipse Public License - v 1.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *         http://www.eclipse.org/legal/epl-v10.html
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package nl.cloudfarming.client.geoviewer.jxmap.map;

import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import javax.swing.JComponent;

/**
 * Top layer on the map, catching all mouse and key events, redirecting them to the map mouse handler.
 *
 * 
 * @author Timon Veenstra
 */
public class MouseAndKeyEventReceiverLayer extends JComponent {

    public MouseAndKeyEventReceiverLayer(MapEventController mapEventController) {
        setOpaque(false);

        addMouseListener(mapEventController);
        addMouseMotionListener(mapEventController);
        addMouseWheelListener(mapEventController);

        addKeyListener(mapEventController);

        // adding mouse adapter to keep focus on this panel. This is needed to
        // direct all mouse input to this panel. For example when dragging
        // mousewheel events aren't received anymore.
        addMouseListener(new MouseAdapter() {
            @Override
            public void mouseReleased(final MouseEvent e) {
                requestFocusInWindow();
            }
        });
    }
}
